package ru.dbotthepony.kstarbound.server.world

import com.google.gson.JsonElement
import com.google.gson.JsonObject
import it.unimi.dsi.fastutil.ints.IntArraySet
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import kotlinx.coroutines.async
import kotlinx.coroutines.future.asCompletableFuture
import kotlinx.coroutines.future.await
import kotlinx.coroutines.runBlocking
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.AABBi
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.math.vector.Vector2i
import ru.dbotthepony.kstarbound.Globals
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.VersionRegistry
import ru.dbotthepony.kstarbound.defs.WarpAction
import ru.dbotthepony.kstarbound.defs.WarpAlias
import ru.dbotthepony.kstarbound.defs.WorldID
import ru.dbotthepony.kstarbound.defs.tile.TileDamage
import ru.dbotthepony.kstarbound.defs.tile.TileDamageResult
import ru.dbotthepony.kstarbound.defs.tile.TileDamageType
import ru.dbotthepony.kstarbound.defs.tile.isEmptyLiquid
import ru.dbotthepony.kstarbound.defs.world.WorldStructure
import ru.dbotthepony.kstarbound.defs.world.WorldTemplate
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.jsonArrayOf
import ru.dbotthepony.kstarbound.network.IPacket
import ru.dbotthepony.kstarbound.world.TileModification
import ru.dbotthepony.kstarbound.network.packets.StepUpdatePacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.PlayerWarpResultPacket
import ru.dbotthepony.kstarbound.network.packets.clientbound.UpdateWorldPropertiesPacket
import ru.dbotthepony.kstarbound.server.StarboundServer
import ru.dbotthepony.kstarbound.server.ServerConnection
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kstarbound.util.BlockableEventLoop
import ru.dbotthepony.kstarbound.util.ActionPacer
import ru.dbotthepony.kstarbound.util.random.random
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.ChunkState
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.WorldGeometry
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.tile.TileEntity
import ru.dbotthepony.kstarbound.world.entities.tile.WorldObject
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import java.util.concurrent.CompletableFuture
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.RejectedExecutionException
import java.util.concurrent.TimeUnit
import java.util.function.Supplier

class ServerWorld private constructor(
	val server: StarboundServer,
	template: WorldTemplate,
	val storage: WorldStorage,
	val worldID: WorldID,
) : World<ServerWorld, ServerChunk>(template) {
	init {
		if (server.isClosed)
			throw RuntimeException()
	}

	val clients = CopyOnWriteArrayList<ServerWorldTracker>()
	val shouldStopOnIdle = worldID !is WorldID.ShipWorld

	private suspend fun doAcceptClient(client: ServerConnection, action: WarpAction?) {
		try {
			isBusy++

			if (clients.any { it.client == client })
				throw IllegalStateException("$client is already in $this")

			if (!client.isConnected)
				throw IllegalStateException("$client disconnected while joining $this")

			val start = if (action is WarpAction.Player)
				clients.firstOrNull { it.client.uuid == action.uuid }?.client?.playerEntity?.position
			else if (action is WarpAction.World)
				action.target.resolve(this)
			else
				playerSpawnPosition

			if (start == null) {
				client.send(PlayerWarpResultPacket(false, action!!, true))
				throw IllegalStateException("Not a valid spawn target: $action")
			}

			if (action != null)
				client.send(PlayerWarpResultPacket(true, action, false))

			client.tracker?.remove("Transiting to new world", false)
			clients.add(ServerWorldTracker(this, client, start))

			//if (worldID is WorldID.Celestial)
				//Registries.dungeons["naturalcave"]?.value?.generate(this@ServerWorld, random(), start.x.toInt(), start.y.toInt(), markSurfaceAndTerrain = false, forcePlacement = true)
		} finally {
			isBusy--
		}
	}

	fun acceptClient(player: ServerConnection, action: WarpAction? = null): CompletableFuture<Unit> {
		check(!eventLoop.isShutdown) { "$this is invalid" }

		try {
			val future = eventLoop.scope.async { doAcceptClient(player, action) }.asCompletableFuture()

			future.exceptionally {
				LOGGER.error("Error while accepting new player into world", it)
			}

			return future
		} catch (err: RejectedExecutionException) {
			return CompletableFuture.failedFuture(err)
		}
	}

	fun saveMetadata() {
		val metadata = MetadataJson(
			playerStart = playerSpawnPosition,
			respawnInWorld = respawnInWorld,
			adjustPlayerStart = adjustPlayerSpawn,
			worldTemplate = if (storage is LegacyWorldStorage) Starbound.legacyJson { template.toJson() } else template.toJson(),
			centralStructure = centralStructure,
			protectedDungeonIds = protectedDungeonIDs,
			worldProperties = copyProperties(),
			spawningEnabled = true
		)

		storage.saveMetadata(WorldStorage.Metadata(geometry, VersionRegistry.make("WorldMetadata", Starbound.gson.toJsonTree(metadata))))
	}

	override val eventLoop = object : BlockableEventLoop("Server World $worldID") {
		init {
			isDaemon = true
		}

		override fun performShutdown() {
			LOGGER.info("Shutting down ${this@ServerWorld}")

			try {
				server.notifyWorldUnloaded(worldID)
			} catch (err: RejectedExecutionException) {
				// do nothing
			}

			chunkMap.chunks().forEach {
				try {
					it.unload()
				} catch (err: Throwable) {
					LOGGER.error("Exception while saving chunk ${it.pos}", err)
				}
			}

			clients.forEach {
				LOGGER.info("Kicking ${it.client.alias()} off dying world")
				it.remove("World shutting down", setReturnWarp = false)
				it.client.enqueueWarp(WarpAlias.Return)
			}

			saveMetadata()
			storage.close()
		}
	}

	init {
		eventLoop.scheduleAtFixedRate(Runnable {
			tick(Starbound.TIMESTEP)
		}, 0L, Starbound.TIMESTEP_NANOS, TimeUnit.NANOSECONDS)

		eventLoop.scheduleWithFixedDelay(Runnable {
			storage.sync()
		}, 10L, 10L, TimeUnit.SECONDS)
	}

	override fun toString(): String {
		return "Server World $worldID"
	}

	private var idleTicks = 0
	private var isBusy = 0

	private val wireProcessor = LegacyWireProcessor(this)

	override val isClient: Boolean
		get() = false

	/**
	 * this method does not block if pacer is null (safe to use with runBlocking {})
	 */
	suspend fun damageTiles(positions: Collection<IStruct2i>, isBackground: Boolean, sourcePosition: Vector2d, damage: TileDamage, source: AbstractEntity? = null, pacer: ActionPacer? = null): TileDamageResult {
		if (damage.amount <= 0.0)
			return TileDamageResult.NONE

		data class TileToDamage(val position: Vector2i, val chunk: ServerChunk?, val roots: MutableList<TileEntity> = ArrayList(0))

		val actualPositions = ObjectArraySet<TileToDamage>()

		for (pos in positions) {
			val wrapped = geometry.wrap(pos)
			val chunk = chunkMap[geometry.chunkFromCell(wrapped)]

			if (chunk != null) {
				val cell = chunk.getCell(wrapped - chunk.pos.tile)

				if (cell.rootSource != null) {
					actualPositions.add(TileToDamage(wrapped, chunk, entityIndex.tileEntitiesAt(geometry.wrap(cell.rootSource!!))))
				} else {
					actualPositions.add(TileToDamage(wrapped, chunk))
				}
			} else {
				actualPositions.add(TileToDamage(wrapped, chunk))
			}
		}

		var topMost = TileDamageResult.NONE
		val toDamageEntities = Object2ObjectArrayMap<TileEntity, ObjectArraySet<Vector2i>>()
		val entityDamageResults = HashMap<Vector2i, TileDamageResult>()

		for ((pos, chunk, roots) in actualPositions) {
			for (root in roots) {
				val posList = toDamageEntities.computeIfAbsent(root, Object2ObjectFunction { ObjectArraySet() })
				posList.add(pos)
				posList.addAll(root.roots)
				posList.add(root.tilePosition)
			}

			if (!isBackground) {
				for (entity in entitiesAtTile(pos)) {
					toDamageEntities.computeIfAbsent(entity, Object2ObjectFunction { ObjectArraySet() }).add(pos)
				}
			}
		}

		val killedEntities = ObjectArraySet<TileEntity>()

		for ((entity, damagePositions) in toDamageEntities.entries) {
			var actualDamage = damage

			for (pos in damagePositions) {
				val cell = getCell(pos)

				if (cell.dungeonId in protectedDungeonIDs) {
					actualDamage = actualDamage.copy(type = TileDamageType.PROTECTED)
					entityDamageResults[pos] = TileDamageResult.PROTECTED
				} else {
					entityDamageResults[pos] = TileDamageResult.NORMAL
				}
			}

			val occupySpaces = entity.occupySpaces.stream()
				.filter { p -> p in damagePositions }
				.toList()

			pacer?.consume(10)
			val broken = entity.damage(occupySpaces, sourcePosition, actualDamage)

			if (source != null && broken) {
				source.receiveMessage("tileEntityBroken", jsonArrayOf(
					damagePositions.firstOrNull { p -> actualPositions.any { it.position == p } } ?: entity.tilePosition,
					entity.type.jsonName,
					(entity as? WorldObject)?.config?.key))
			}

			if (broken) {
				killedEntities.add(entity)
				entityDamageResults.keys.removeAll(damagePositions)
			}
		}

		actualPositions.removeIf {
			it.roots.removeIf { it in killedEntities }
			it.roots.isNotEmpty()
		}

		for ((pos, chunk) in actualPositions) {
			var actualDamage = damage
			val tileEntityResult = entityDamageResults[pos] ?: TileDamageResult.NONE
			val getCell = chunk?.getCell(pos - chunk.pos.tile)

			if (getCell != null) {
				if (getCell.dungeonId in protectedDungeonIDs) {
					actualDamage = actualDamage.copy(type = TileDamageType.PROTECTED)
				}
			}

			// Penetrating damage should carry through to the blocks behind this
			// entity.
			if (tileEntityResult == TileDamageResult.NONE || actualDamage.type.isPenetrating) {
				chunk ?: continue
				pacer?.consume()
				val (result, health, tile) = chunk.damageTile(pos - chunk.pos.tile, isBackground, sourcePosition, actualDamage, source)
				topMost = topMost.coerceAtLeast(result)

				if (source != null && health?.isDead == true) {
					source.receiveMessage("tileBroken", jsonArrayOf(
						pos, if (isBackground) "background" else "foreground",
						tile!!.tile(isBackground).material.id ?: tile.tile(isBackground).material.key, // TODO: explicit string identifiers support
						tile.dungeonId,
						health.isHarvested
					))
				}
			}

			topMost = topMost.coerceAtLeast(tileEntityResult)
		}

		return topMost
	}

	override fun applyTileModifications(modifications: Collection<Pair<Vector2i, TileModification>>, allowEntityOverlap: Boolean, ignoreTileProtection: Boolean): List<Pair<Vector2i, TileModification>> {
		return runBlocking { applyTileModifications(modifications, allowEntityOverlap, ignoreTileProtection, null) }
	}

	suspend fun applyTileModifications(modifications: Collection<Pair<Vector2i, TileModification>>, allowEntityOverlap: Boolean, ignoreTileProtection: Boolean = false, pacer: ActionPacer?): List<Pair<Vector2i, TileModification>> {
		val unapplied = ArrayList(modifications)
		var size: Int

		do {
			size = unapplied.size
			val itr = unapplied.iterator()

			for ((pos, modification) in itr) {
				val cell = getCell(pos)

				if (!ignoreTileProtection && cell.dungeonId in protectedDungeonIDs)
					continue

				if (modification.allowed(this, pos, allowEntityOverlap)) {
					pacer?.consume()
					modification.apply(this, pos, allowEntityOverlap)
					itr.remove()
				}
			}
		} while(unapplied.isNotEmpty() && size != unapplied.size)

		return unapplied
	}

	override fun tick(delta: Double) {
		try {
			if (clients.isEmpty() && isBusy <= 0) {
				idleTicks++
			} else {
				idleTicks = 0
			}

			if (idleTicks >= 600) {
				if (shouldStopOnIdle) {
					eventLoop.shutdown()
				}

				return
			}

			if (!Globals.universeServer.useNewWireProcessing) {
				wireProcessor.tick()
			}

			super.tick(delta)

			val packet = StepUpdatePacket(ticks)

			clients.forEach {
				it.send(packet)

				try {
					it.tick()
				} catch (err: Throwable) {
					LOGGER.error("Exception while ticking player ${it.client}", err)
					//it.disconnect("Exception while ticking player: $err")
				}
			}
		} catch (err: Throwable) {
			LOGGER.fatal("Exception in world tick loop", err)
			eventLoop.shutdown()

			if (worldID is WorldID.ShipWorld) {
				server.channels.connectionByUUID(worldID.uuid)?.disconnect("Exception in shipworld tick loop: $err")
			}
		}
	}

	override fun broadcast(packet: IPacket) {
		clients.forEach {
			it.send(packet)
		}
	}

	var isInPreparation = false
		private set

	private suspend fun prepare0(firstTime: Boolean) {
		try {
			isInPreparation = true
			isBusy++

			if (firstTime) {
				val random = random(template.seed)
				var currentDungeonID = 0

				// primary dungeons must be generated sequentially since they can get very large
				// and to avoid dungeons colliding with each other, we must generate them first, one by one
				for (dungeon in template.gatherDungeons()) {
					var spawnDungeonRetries = Globals.worldServer.spawnDungeonRetries

					do {
						var x = dungeon.baseX

						if (dungeon.xVariance != 0)
							x += random.nextInt(dungeon.xVariance)

						x = geometry.x.cell(x)

						LOGGER.info("Trying to place dungeon ${dungeon.dungeon.key} at $x, ${dungeon.baseHeight}...")

						val dungeonWorld = try {
							dungeon.dungeon.value.generate(this, random(random.nextLong()), x, dungeon.baseHeight, dungeon.blendWithTerrain, dungeon.forcePlacement, dungeonID = currentDungeonID).await()
						} catch (err: Throwable) {
							LOGGER.error("Exception while placing dungeon ${dungeon.dungeon.key} at $x, ${dungeon.baseHeight}", err)
							// continue
							break
						}

						if (dungeonWorld.hasGenerated) {
							LOGGER.info("Placed dungeon ${dungeon.dungeon.key} at $x, ${dungeon.baseHeight}")

							if (dungeon.dungeon.value.metadata.protected) {
								protectedDungeonIDs.add(currentDungeonID)
							}

							if (dungeon.dungeon.value.metadata.gravity != null) {
								// TODO: set gravity here
							}

							if (dungeon.dungeon.value.metadata.breathable != null) {
								// TODO: set "breathable" here
							}

							currentDungeonID++
							break
						}
					} while (--spawnDungeonRetries > 0)
				}
			}

			if (playerSpawnPosition == Vector2d.ZERO) {
				playerSpawnPosition = findPlayerStart()
			}
		} finally {
			isBusy--
			isInPreparation = false
		}
	}

	// done as completable future because we must do
	// everything inside our own thread, not anywhere else
	// This way, external callers can properly wait for preparations to complete
	fun prepare(firstTime: Boolean): CompletableFuture<*> {
		return eventLoop.scope.async { prepare0(firstTime) }.asCompletableFuture()
	}

	private suspend fun findPlayerStart(hint: Vector2d? = null): Vector2d {
		//if (true) {
		//	return Vector2d(10.0, template.surfaceLevel().toDouble())
		//}

		val tickets = ArrayList<ServerChunk.ITicket>()
		val random = if (hint == null) random(template.seed) else random()

		try {
			LOGGER.info("Trying to find player spawn position...")
			var pos = hint ?: CompletableFuture.supplyAsync(Supplier { template.findSensiblePlayerStart(random) }, Starbound.EXECUTOR).await() ?: Vector2d(0.0, template.surfaceLevel().toDouble())
			var previous = pos
			LOGGER.info("Trying to find player spawn position near $pos...")

			for (t in 0 until Globals.worldServer.playerStartRegionMaximumTries) {
				var foundGround = false

				// First go downward until we collide with terrain
				for (i in 0 until Globals.worldServer.playerStartRegionMaximumVerticalSearch) {
					val spawnRect = AABB(
						Vector2d(pos.x - Globals.worldServer.playerStartRegionSize.x / 2, pos.y),
						Vector2d(pos.x + Globals.worldServer.playerStartRegionSize.x / 2, pos.y + Globals.worldServer.playerStartRegionSize.y),
					)

					val region = permanentChunkTicket(spawnRect).await()
					tickets.addAll(region)
					region.forEach { it.chunk.await() }

					foundGround = anyCellSatisfies(spawnRect) { tx, ty, tcell ->
						tcell.foreground.material.value.collisionKind != CollisionType.NONE
					}

					if (foundGround) {
						break
					} else {
						pos += Vector2d.NEGATIVE_Y
					}
				}

				if (foundGround) {
					// Then go up until our spawn region is no longer in the terrain, but bail
					// out and try again if we can't signal the region or we are stuck in a
					// dungeon.

					for (i in 0 until Globals.worldServer.playerStartRegionMaximumVerticalSearch) {
						if (!getCell(pos.x.toInt(), pos.y.toInt()).liquid.state.isEmptyLiquid) {
							break
						}

						val spawnRect = AABB(
							Vector2d(pos.x - Globals.worldServer.playerStartRegionSize.x / 2, pos.y),
							Vector2d(pos.x + Globals.worldServer.playerStartRegionSize.x / 2, pos.y + Globals.worldServer.playerStartRegionSize.y),
						)

						val region = permanentChunkTicket(spawnRect).await()
						tickets.addAll(region)
						region.forEach { it.chunk.await() }

						if (!anyCellSatisfies(spawnRect) { tx, ty, tcell -> tcell.foreground.material.value.collisionKind != CollisionType.NONE } && spawnRect.maxs.y < geometry.size.y) {
							LOGGER.info("Found appropriate spawn position at $pos")
							return pos
						}

						pos += Vector2d.POSITIVE_Y
					}
				}

				pos = CompletableFuture.supplyAsync(Supplier { template.findSensiblePlayerStart(random) }, Starbound.EXECUTOR).await() ?: Vector2d(0.0, template.surfaceLevel().toDouble())

				if (previous != pos) {
					LOGGER.info("Still trying to find player spawn position near $pos...")
					previous = pos
				} else {
					break
				}
			}

			LOGGER.warn("Unable to find proper player start location, will use $pos")
			return pos
		} finally {
			tickets.forEach { it.cancel() }
		}
	}

	override fun setProperty0(key: String, value: JsonElement) {
		broadcast(UpdateWorldPropertiesPacket(JsonObject().apply { add(key, value) }))
	}

	override fun chunkFactory(pos: ChunkPos): ServerChunk {
		return ServerChunk(this, pos)
	}

	fun permanentChunkTicket(pos: ChunkPos, target: ChunkState = ChunkState.FULL): CompletableFuture<ServerChunk.ITicket?> {
		return eventLoop.supplyAsync { chunkMap.compute(pos)?.permanentTicket(target) }
	}

	fun permanentChunkTicket(region: AABBi, target: ChunkState = ChunkState.FULL): CompletableFuture<List<ServerChunk.ITicket>> {
		return eventLoop.supplyAsync { geometry.region2Chunks(region).mapNotNull { permanentChunkTicket(it, target).get() } }
	}

	fun permanentChunkTicket(region: AABB, target: ChunkState = ChunkState.FULL): CompletableFuture<List<ServerChunk.ITicket>> {
		return eventLoop.supplyAsync { geometry.region2Chunks(region).mapNotNull { permanentChunkTicket(it, target).get() } }
	}

	fun temporaryChunkTicket(pos: ChunkPos, time: Int, target: ChunkState = ChunkState.FULL): CompletableFuture<ServerChunk.ITimedTicket?> {
		return eventLoop.supplyAsync { chunkMap.compute(pos)?.temporaryTicket(time, target) }
	}

	fun temporaryChunkTicket(region: AABBi, time: Int, target: ChunkState = ChunkState.FULL): CompletableFuture<List<ServerChunk.ITimedTicket>> {
		require(time >= 0) { "Invalid ticket time: $time" }

		return eventLoop.supplyAsync { geometry.region2Chunks(region).mapNotNull { temporaryChunkTicket(it, time, target).get() } }
	}

	fun temporaryChunkTicket(region: AABB, time: Int, target: ChunkState = ChunkState.FULL): CompletableFuture<List<ServerChunk.ITimedTicket>> {
		require(time >= 0) { "Invalid ticket time: $time" }

		return eventLoop.supplyAsync { geometry.region2Chunks(region).mapNotNull { temporaryChunkTicket(it, time, target).get() } }
	}

	@JsonFactory
	data class MetadataJson(
		val playerStart: Vector2d,
		val respawnInWorld: Boolean,
		val adjustPlayerStart: Boolean,
		val worldTemplate: JsonObject,
		val centralStructure: WorldStructure,
		val protectedDungeonIds: IntArraySet,
		val worldProperties: JsonObject,
		val spawningEnabled: Boolean
	)

	companion object {
		private val LOGGER = LogManager.getLogger()

		fun create(server: StarboundServer, template: WorldTemplate, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): ServerWorld {
			return ServerWorld(server, template, storage, worldID)
		}

		fun create(server: StarboundServer, geometry: WorldGeometry, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): ServerWorld {
			return ServerWorld(server, WorldTemplate(geometry), storage, worldID)
		}

		fun create(server: StarboundServer, metadata: WorldStorage.Metadata, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): ServerWorld {
			return AssetPathStack("/") { _ ->
				val meta = Starbound.gson.fromJson(metadata.data.content, MetadataJson::class.java)

				val world = ServerWorld(server, WorldTemplate.fromJson(meta.worldTemplate), storage, worldID)
				world.playerSpawnPosition = meta.playerStart
				world.respawnInWorld = meta.respawnInWorld
				world.adjustPlayerSpawn = meta.adjustPlayerStart
				world.centralStructure = meta.centralStructure

				for ((k, v) in meta.worldProperties.entrySet()) {
					world.setProperty(k, v)
				}

				world.protectedDungeonIDs.addAll(meta.protectedDungeonIds)
				world
			}
		}

		fun load(server: StarboundServer, storage: WorldStorage, worldID: WorldID = WorldID.Limbo): CompletableFuture<ServerWorld> {
			LOGGER.info("Attempting to load world at $worldID")

			return storage.loadMetadata().thenApply {
				LOGGER.info("Loading world at $worldID")
				AssetPathStack("/") { _ ->
					val meta = it.map { Starbound.gson.fromJson(it.data.content, MetadataJson::class.java) }.orThrow { NoSuchElementException("No world metadata is present") }

					val world = ServerWorld(server, WorldTemplate.fromJson(meta.worldTemplate), storage, worldID)
					world.playerSpawnPosition = meta.playerStart
					world.respawnInWorld = meta.respawnInWorld
					world.adjustPlayerSpawn = meta.adjustPlayerStart
					world.centralStructure = meta.centralStructure

					for ((k, v) in meta.worldProperties.entrySet()) {
						world.setProperty(k, v)
					}

					world.protectedDungeonIDs.addAll(meta.protectedDungeonIds)
					world
				}
			}.also {
				it.exceptionally {
					LOGGER.error("Error while instancing world $worldID", it)
					null
				}
			}
		}
	}
}
